#!/usr/bin/env python3

import os
import numpy as np
import math
from pandas import Series
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib.gridspec import GridSpec
from datetime import datetime, timezone
from matplotlib.font_manager import FontProperties
from configparser import ConfigParser
from .forward_module.spheroid import spheroid
from .forward_module.tamuDUST import ih

# current_dir = os.path.dirname(os.path.abspath(__file__))


def get_BOREAL_version(setup_dir):
    setup_file = os.path.join(setup_dir, 'setup.cfg')
    config = ConfigParser()
    config.read(setup_file)
    return config['metadata']['version']


#setup_directory = os.path.join(current_dir, '..')
#BOREAL_v = get_BOREAL_version(setup_directory)


class BorealException(Exception):
    pass


# construct the second difference operator, length of the objective vector - 2 !!!
def construct_U2(length):
    k_array = np.zeros((length - 2, length))
    for h_i in range(length - 2):
        k_array[h_i, h_i] = 1
        k_array[h_i, h_i + 1] = -2
        k_array[h_i, h_i + 2] = 1
    zero = np.zeros((length - 2, 2))
    array_jacob = np.concatenate((k_array, zero), axis=1)
    return k_array, array_jacob


def derivative(func, xx, args=None, dx=1e-7):
    """
    calculate the Jacobian matrix and the value vector of the vector-function 'f' at 'x'
    :param args: other parameters in f(x), a dictionary
    :param func: objective function, callable, output a 1-D array as the objective vector
    :param xx: the variety vector of 'f', 1-D array (ndarray or list)
    :param dx: delta
    :return: J(x)-ndarray and f(x)-ndarray
    """
    if args is None:
        args = {}
    xx = list(xx)
    y = func(xx, **args)
    try:
        Jacob_T = np.empty((len(xx), len(y)))
    except TypeError:
        Jacob_T = np.empty((len(xx), 1))
    for line_num in range(Jacob_T.shape[0]):
        x_new = xx[:]
        x_new[line_num] += dx
        y_new = func(x_new, **args)
        Jacob_T[line_num] = (y_new - y) / dx
    Jacob = Jacob_T.T
    return Jacob, y


# get the number of modes of a particle size distribution
def get_mode_num(psd):
    # locate local maxima
    diff_v = psd[1:] - psd[: -1]
    vmax_num = 0  # Note: they are the indices of local maxima
    for ii in range(1, len(psd) - 1):
        if (diff_v[ii] < 0) & (diff_v[ii - 1] > 0):
            vmax_num += 1
    return vmax_num


def Vt_Reff(grid, vi):
    """
    calculate volume concentration, effective radius from a VSD
    :return:
    """
    ln_grid = np.log(grid)
    concentrate_vol = np.trapz(vi, ln_grid)
    effective_r = concentrate_vol / np.trapz(vi / grid, ln_grid)
    return concentrate_vol, effective_r


def statistic_quant(data):
    """
    get mean and std of an array. If the array is 2d of size (m, n), the statistics is conducted
    along the m-dimension, resulting in a 1d array of size n.
    :param data: numpy or list data
    :return:
    """
    nd_data = np.array(data)
    if len(nd_data.shape) == 1:
        data_mean = np.mean(nd_data)
        data_std = np.std(nd_data)
    else:
        data_mean = np.mean(nd_data, axis=0)
        data_std = np.std(nd_data, axis=0)
    return data_mean, data_std


def get_default_confg(aerosol_type):
    """
    get default configuration parameters for inversion()
    :return:
    """
    # input configuration
    na, sigma_na = 1.5, 0.1
    ka_dict = {'absorbing': 0.015, 'dust': 0.005, 'non-absorbing': 0.005}  # 'ss' for sea salt
    sigma_ka_dict = {'absorbing': 0.01, 'dust': 0.005, 'non-absorbing': 0.005}
    if aerosol_type == 'bba':
        ka = ka_dict['absorbing']
        sigma_ka = sigma_ka_dict['absorbing']
    elif aerosol_type == 'dust':
        ka = ka_dict['dust']
        sigma_ka = sigma_ka_dict['dust']
    else:
        ka = ka_dict['non-absorbing']
        sigma_ka = sigma_ka_dict['non-absorbing']
    rmin_min = 0.05
    rmin_max = 0.3
    rmax_min = 1.5
    rmax_max = 15
    r_regrid_num = 8

    epsilon2 = 2.5
    # epsilon3 = sigma_na / na
    epsilon_delta_x = 2.54  # (ln(xmax) - ln(xmin)) / 2
    epsilon_delta_n = 0.07  # (ln1.6 - ln1.4) / 2
    epsilon_delta_k = 2.3  # (ln0.1 - ln0.001) / 2 # scaling matrix

    return {'na': na, 'sigma_na': sigma_na, 'ka': ka, 'sigma_ka': sigma_ka, 'rmin_min': rmin_min, 'rmin_max': rmin_max,
            'rmax_min': rmax_min, 'rmax_max': rmax_max, 'r_regrid_num': r_regrid_num, 'epsilon_delta_x': epsilon_delta_x,
            'epsilon_delta_n': epsilon_delta_n, 'epsilon_delta_k': epsilon_delta_k, 'epsilon2': epsilon2}


def get_default_measerr():
    """
    get default measurement error
    :return: a dict
    """
    return {'ext_err': 0.1, 'bac_err': 0.1, 'depol_err': 0.1, 'lr_err': 0.15}


def __do_iter(beta, Y, fwd_func, arg_fwd, D, U, J2, J3, J4, SI, r_regrid_num, res_vec):
    """
    do iteration
    :param Y: list, ln[y1, y2, y3, y4]
    :param fwd_func: callable, function for forward calculation, called by derivative()
    :param beta: list or ndarray, logarithm of the initial value of the state parameter vector
    :param res_vec: ndarray, measurement error vector, relative error. The size is equal to Y[0].
    :param arg_fwd: dict, arguments of fwd_model.get_opt other than the state parameters
    :return:
    """
    # the first iteration, i.e., initialization
    Iter_num = 0
    beta_exp = np.exp(beta)
    # calculate the Jacobian matrix
    J1, f1 = derivative(func=fwd_func, xx=beta_exp, args=arg_fwd)  # array
    U1 = J1 * beta_exp / f1[:, np.newaxis]
    # calculate fl
    Xi = np.mat(beta[:-2]).T
    F1 = np.log(f1)
    F2 = np.array(U * Xi).ravel()
    F3 = beta[-2]  # f3 = ni
    F4 = beta[-1]  # f4 = ki
    Fl = [F1, F2, F3, F4]
    J = [np.mat(U1), np.mat(J2), np.mat(J3), np.mat(J4)]
    num_gener_meas = F1.size + F2.size + 2
    c = (num_gener_meas - r_regrid_num - 2)
    sum3 = 0
    for l in range(len(Fl)):
        deltaY_l = np.mat(Y[l] - Fl[l]).T
        sum3 += deltaY_l.T * SI[l] * deltaY_l
    chi_square = sum3[0, 0]
    damp_fac = 2 / c * chi_square
    enlarge_factor = 2

    # the following iterations
    while Iter_num <= 30:
        sum1, sum2 = 0, 0
        for l in range(len(Fl)):
            alpha_l = J[l].T * SI[l]
            deltaY_l = np.mat(Y[l] - Fl[l]).T
            sum1 += alpha_l * J[l]
            sum2 += alpha_l * deltaY_l
        G = sum1 + damp_fac * D
        b = sum2
        GI = G.I
        delta_beta = np.array(GI * b).ravel()

        beta_after = delta_beta + beta
        beta_exp_after = np.exp(beta_after)
        # calculate the Jacobian matrix
        J1_after, f1_after = derivative(func=fwd_func, xx=beta_exp_after, args=arg_fwd)  # array
        U1_after = J1_after * beta_exp_after / f1_after[:, np.newaxis]
        # calculate fl
        Xi_after = np.mat(beta_after[:-2]).T
        F1_after = np.log(f1_after)
        F2_after = np.array(U * Xi_after).ravel()
        F3_after = beta_after[-2]  # f3 = ni
        F4_after = beta_after[-1]  # f4 = ki
        Fl_after = [F1_after, F2_after, F3_after, F4_after]
        J_after = [np.mat(U1_after), np.mat(J2), np.mat(J3), np.mat(J4)]
        sum3_after = 0
        for l in range(len(Fl_after)):
            deltaY_l = np.mat(Y[l] - Fl_after[l]).T
            sum3_after += deltaY_l.T * SI[l] * deltaY_l
        chi_square_after = sum3_after[0, 0]
        Iter_num += 1
        if chi_square_after >= chi_square:
            damp_fac *= enlarge_factor
            enlarge_factor *= 2
        else:
            chi_square = chi_square_after
            damp_fac = max(2 / c * chi_square, damp_fac / 3)
            beta = beta_after
            Fl = Fl_after
            J = J_after
            enlarge_factor = 2

        deltaY_1 = np.mat(Y[0] - Fl_after[0]).T
        delta_Y1_array = np.array(deltaY_1).ravel()
        rel_err = np.fabs(np.exp(-delta_Y1_array) - 1)
        # stop = (Iter_num >= 30) | (np.all(res_vec - rel_err) & (chi_square_after < c))
        stop = np.all(res_vec - rel_err) & (chi_square < c)
        if stop:
            break

    e_for_measurement = np.linalg.norm(rel_err) / rel_err.size ** 0.5 * 100
    return beta, Iter_num, Fl[0], chi_square, e_for_measurement


def filt_inval_meas(name, dict_data):
    """
    filter invalid measurements (NaN, float less than zero are considered as 'invalid')
    :param dict_data:
    :param name:
    :return: a Series type if the input data are valid
    """
    for dict_item in dict_data.items():
        if dict_item[1] < 0:
            raise BorealException('%s at %s nm is negative.' % (name, dict_item[0]))
    seri_data = Series(dict_data)
    return seri_data


def check_meas_and_err(name, meas, err, err_default):
    if not len(meas):
        raise BorealException('%s measurements cannot be empty.' % name)
    meas_series = filt_inval_meas(name, meas)
    if (err is None) or (not len(err)):
        err = err_default
    err_series = filt_inval_meas(name+'_err', err)
    if len(set(err_series.index) & set(meas_series.index)):
        err_series = err_series[meas_series.index]
    else:
        raise BorealException('For %s: the error wavelengths is not consistent with the measurement wavelengths.' % name)
    meas_wave = meas_series.index.values.astype(float) * 0.001
    meas_series.index = meas_wave
    err_series.index = meas_wave
    meas_wave.sort()

    return meas_series.reindex(meas_wave), err_series.reindex(meas_wave), meas_wave


def inversion(ext, bac, aero_type, model, depol=None, ext_err=None, bac_err=None, depol_err=None, config=None):
    """
    main function to implement the inverse procedure
    :param ext: dict, spectral ext. coef., the keys (str) are wavelength in nm, the values are corresponding measurements (float) in 1/Mm
    :param bac: dict, spectral bac. coef., the keys (str) are wavelength in nm, the values are corresponding measurements (float) in 1/(Mm*sr)
    :param depol: dict or None (default=None), spectral depol., the keys (str) are wavelength in nm, the values are corresponding measurements (float) (unit of 1)
    :param aero_type: str, 'dust', 'bba', 'urban' or 'ss' (sea salt)
    :param model: str, 'sphere', 'spheroid' or 'ih', forward model (scattering model) used in the inversion
    :param ext_err: dict or None (default=None), maximum measurement error in ext (three times of measurement std). None for default values.
    :param bac_err: dict or None, measurement error in bac, having the same form with bac. If None, get default configuration. By default, None
    :param depol_err: dict or None, measurement error in depol, having the same form with depol. If None, get default configuration. By default, None
    :param config: dict, other configuration parameters
    :return: two dicts: Retrieval and Fit
    """
    # time info:
    now_utc = datetime.now(tz=timezone.utc)
    now_local = now_utc.astimezone()  # covert timezone to the system default.
    strfmd_now_utc = now_utc.strftime('%Y-%m-%dT%H:%M:%S')
    strfmd_now_local = now_local.strftime('%Y-%m-%dT%H:%M:%S')
    str_time = "Processed at:\n%s (%s time)\n%s (%s time)" % (strfmd_now_utc, now_utc.tzname(), strfmd_now_local, now_local.tzname())

    # pre-treatment
    # check model name and aerosol type
    if aero_type not in ['dust', 'bba', 'urban', 'ss']:
        raise BorealException('The aerosol type must be one of the following: %s' % 'absorbing, non-absorbing, dust')
    if aero_type == 'dust':
        mI_dusttype = True
    else:
        mI_dusttype = False

    if model == 'ih':
        r_grid_org = ih.r_grid_org
        get_opt = ih.get_opt
    elif (model == 'spheroid') | (model == 'sphere'):
        r_grid_org = spheroid.r_grid_org
        get_opt = spheroid.get_opt
    else:
        raise BorealException('The model name must be one of the following: %s' % 'sphere, spheroid, ih')

    # check ext & ext_err; bac & bac_err
    err_default_ext_and_bac = {'355': 0.1, '532': 0.1, '1064': 0.2}
    err_default_depol = {'355': 0.15, '532': 0.15, '1064': 0.15}
    ext_series, ext_err_series, ext_wave = check_meas_and_err('Extinction', ext, ext_err, err_default=err_default_ext_and_bac)
    bac_series, bac_err_series, bac_wave = check_meas_and_err('Backscattering', bac, bac_err, err_default=err_default_ext_and_bac)

    # check depol
    # check availability of depol meas
    if (depol is None) or (not len(depol)):
        depol_wave = None
        depol_series = {}
        depol_err_series = {}
    else:
        depol_series, depol_err_series, depol_wave = check_meas_and_err('Depolarization', depol, depol_err, err_default=err_default_depol)

    # determine inversion dataset
    if (model == 'sphere') | (depol_wave is None):
        y1 = np.concatenate((ext_series.values, bac_series.values))
        epsilon1 = np.concatenate((ext_err_series.values, bac_err_series.values)) / 3
        print('The %s model is used to invert alpha at %s um, and beta at %s um' % (model, ', '.join([str(i) for i in ext_series.index]), ', '.join([str(i) for i in bac_series.index])))
    else:
        y1 = np.concatenate((ext_series.values, bac_series.values, depol_series.values))
        epsilon1 = np.concatenate((ext_err_series.values, bac_err_series.values, depol_err_series.values)) / 3
        print('The %s model is used to invert alpha at %s um, beta at %s um, and delta at %s um' % (
              model, ', '.join([str(i) for i in ext_series.index]), ', '.join([str(i) for i in bac_series.index]), ', '.join([str(i) for i in depol_series.index])))

    args0 = {'wave_ext': ext_wave, 'wave_bac': bac_wave, 'wave_depol': depol_wave, 'mI_dusttype': mI_dusttype, 'model': model, 'only_fwd': False}

    # check if the length of y1 >= 3, otherwise, c <= 0 (underdetermined system, measurements < state parameters)
    if len(y1) < 3:
        raise BorealException('Underdetermined inverse system: the total number of the input measurements is less than 3')

    # calculate lidar ratio
    lr_series = ext_series / bac_series

    # calculate lidar ratio error (maximum relative error)
    partial_lr_ext = 1 / bac_series
    partial_lr_bac = - ext_series / bac_series ** 2
    lr_err_term1 = (ext_err_series / lr_series * partial_lr_ext * ext_err_series / 3) ** 2
    lr_err_term2 = (bac_series / lr_series * partial_lr_bac * bac_err_series / 3) ** 2
    lr_err_series = (lr_err_term1 + lr_err_term2) ** 0.5 * 3
    lr_series = lr_series.dropna()
    lr_err_series = lr_err_series.dropna()

    if config is None:
        config = get_default_confg(aerosol_type=aero_type)

    rmin_start = np.argmin(np.fabs(r_grid_org - config['rmin_min']))
    rmin_end = np.argmin(np.fabs(r_grid_org - config['rmin_max']))
    rmax_start = np.argmin(np.fabs(r_grid_org - config['rmax_min']))
    rmax_end = np.argmin(np.fabs(r_grid_org - config['rmax_max']))
    ka, sigma_ka = config['ka'], config['sigma_ka']

    D = np.mat(np.diag(np.hstack((np.ones(config['r_regrid_num']) / config['epsilon_delta_x'] ** 2,
                                  1 / config['epsilon_delta_n'] ** 2, 1 / config['epsilon_delta_k'] ** 2))))
    U, J2 = construct_U2(length=config['r_regrid_num'])  # array
    J3 = np.hstack((np.zeros(config['r_regrid_num']), [1, 0]))
    J4 = np.hstack((np.zeros(config['r_regrid_num']), [0, 1]))
    Y2 = np.zeros(config['r_regrid_num'] - 2)

    S2 = np.diag(config['epsilon2'] ** 2 * np.ones(config['r_regrid_num'] - 2))
    epsilon3 = config['sigma_na'] / config['na']
    S3 = np.diag(epsilon3 ** 2 * np.ones(1))  # na cov
    config2 = {'D': D, 'U': U, 'J2': J2, 'J3': J3, 'J4': J4}

    r_grid_eff = r_grid_org[rmin_start: rmax_end + 1]
    ln_r_grid_eff = np.log(r_grid_eff)  # both the kernel function and dV_dlnr should be interpolated in the log-scale
    epsilon4 = math.log((1 + (1 + 4 * (sigma_ka / ka) ** 2) ** 0.5) / 2) ** 0.5
    S4 = np.diag(epsilon4 ** 2 * np.ones(1))  # ka cov
    Y = [np.log(y1), Y2, math.log(config['na']), math.log(ka)]
    rms_epsilon1 = np.linalg.norm(epsilon1) / epsilon1.size ** 0.5 * 100
    config2['res_vec'] = epsilon1 * 3
    scale_fac_for_epsilon1 = 1

    while True:
        epsilon1_enlarge = scale_fac_for_epsilon1 * epsilon1
        S1 = np.diag(epsilon1_enlarge ** 2)  # measurement cov
        SI = [np.mat(S1).I, np.mat(S2).I, np.mat(S3).I, np.mat(S4).I]
        config2['SI'] = SI

        # inversion for all the inversion windows
        Solution_org = []
        Solution_org1 = []
        VSD_IS = []  # list of individual solutions before the selection criteria are applied
        for rmin_index in range(rmin_start, rmin_end + 1):
            for rmax_index in range(rmax_start, rmax_end + 1):
                if rmax_index - rmin_index < 14:  # #####################
                    continue
                r_grid_cutforinv = r_grid_org[rmin_index: rmax_index + 1]
                psd_unif = np.ones(config['r_regrid_num'])
                ln_r_grid_inv = np.linspace(math.log(r_grid_cutforinv[0]), math.log(r_grid_cutforinv[-1]), num=config['r_regrid_num'])
                r_grid_inv = np.exp(ln_r_grid_inv)
                args1 = {'r_grid': r_grid_cutforinv, 'lnr_grid_redu': ln_r_grid_inv}
                args = {**args0, **args1}
                # initialize VSD as a uniform distribution calculated from ext532 (or ext355)
                ext_unif = get_opt(np.hstack((psd_unif, config['na'], ka)), **args)[: len(ext_wave)]
                for wave_unif, ind_in_ext_wave in zip([0.532, 0.355, 1.064], [1, 0, 2]):
                    if wave_unif in ext_wave:
                        VSD_initial = 0.05 * ext_series[wave_unif] / ext_unif[ind_in_ext_wave] * np.ones(config['r_regrid_num'])
                        break

                State_param0 = np.log(np.hstack((VSD_initial, config['na'], ka)))  # the reason why don't put it outside the loop is that the iteration in "__do_iter" will change its value.
                # Thus it needs initialisation each time.
                try:
                    log_xi, iter_num, Fl_after0, cost_func, e_for_measurement = __do_iter(State_param0, Y, fwd_func=get_opt, r_regrid_num=config['r_regrid_num'], arg_fwd=args, **config2)
                except (IndexError, ValueError):
                    continue
                xi = np.exp(log_xi)
                # y1_recal = np.exp(-delta_Y1_array + self.Y[0])
                # first filter: threshold of geometric variance of the retrieved size distribution
                vsd_is = xi[:-2]
                VSD_IS.append(np.vstack((r_grid_inv, vsd_is)))  # collect every individual solution (r_grid_inv, vsd_is)
                vi_max = vsd_is.max()
                threshold1 = 0.5 * vi_max
                # threshold1 = 0.7 * vi_max  # historical used
                threshold2 = 0.05 * vi_max
                args['r_grid_inv'] = r_grid_inv
                if (((vsd_is[0] < vsd_is[1]) & (vsd_is[0] <= threshold1)) | ((vsd_is[0] >= vsd_is[1]) & (vsd_is[0] <= threshold2))) & \
                        (((vsd_is[-1] < vsd_is[-2]) & (vsd_is[-1] <= threshold1)) | ((vsd_is[-1] >= vsd_is[-2]) & (vsd_is[-1] <= threshold2))):
                    mode_params = get_mode_num(vsd_is)
                    if mode_params <= 2:
                        Solution_org.append((e_for_measurement, args, xi, (rmin_index, rmax_index)))
                    else:
                        Solution_org1.append((e_for_measurement, args, xi, (rmin_index, rmax_index)))
                else:
                    Solution_org1.append((e_for_measurement, args, xi, (rmin_index, rmax_index)))
        if len(Solution_org1) or len(Solution_org):
            break
        elif scale_fac_for_epsilon1 < 4:
            scale_fac_for_epsilon1 += 1
            # print('Warning: Overflow happened during the iteration. Enlarge epsilon1 by a factor of %d and try again' % scale_fac_for_epsilon1)
        else:
            rtv_message = 'Warning: Overflow still happened even if epsilon1 was enlarged by a factor of %d. Not any individual solutions are found' % scale_fac_for_epsilon1
            print(rtv_message)
            rtv_flag = 0
            retrieval = {'rtv_message': rtv_message, 'rtv_flag': rtv_flag}
            fit = {}
            return retrieval, fit

    if len(Solution_org):
        # modification on 6 April: sort the discrepancy instead of using np.percentile
        Solution_org.sort()
        len_Solution = len(Solution_org)
        Solution_array = np.array(Solution_org, dtype=object)
        rtv_message = 'Individual solutions satisfying the constraints on modes and fringe size bins are found'
        rtv_flag = 1
        print(rtv_message)
    else:
        Solution_org1.sort()
        len_Solution = len(Solution_org1)
        Solution_array = np.array(Solution_org1, dtype=object)
        rtv_message = 'Warning: individual solutions are found, but none satisfies the constraints on modes or fringe size bins'
        rtv_flag = 2
        print(rtv_message)

    if len_Solution == 1:
        Solution_cut = Solution_array
    else:
        if len_Solution < 5:
            select_num = 1
        else:
            select_frac = 0.2
            select_num = int(len_Solution * select_frac)
        # bool_index = np.hstack((np.ones(1, dtype=int), np.zeros(len_Solution_org - 1, dtype=int)))  # Note!! The dtype should be 'bool' rather than 'int', because 'bool' indexing is in fact a 'fancy' indexing where the rows whose index is True will be obtained.
        # if the type of the indexing list is 'int', only copies of row 0 and row 1 are obtained!! In this case, np.nonzero() must be used to indicate the indices of '1' elements. See the example below.
        bool_index = np.hstack((np.ones(select_num, dtype=bool), np.zeros(len_Solution - select_num, dtype=bool)))
        # Solution_cut = Solution_array[np.nonzero(bool_index | (Solution_array[:, 0] < rms_epsilon1))]  # example of the use of np.nonzero() if the type of bool_index is 'int'
        Solution_cut = Solution_array[bool_index | (Solution_array[:, 0] < rms_epsilon1)]
    Num_of_solutions = Solution_cut.shape[0]
    vsd_coll, n_coll, k_coll, vt_coll, reff_coll = [], [], [], [], []
    ext_full_cal_coll, bac_full_cal_coll, ssa_full_coll, depol_full_cal_coll, lr_full_cal_coll = [], [], [], [], []
    residual_coll = []
    for i in range(Solution_cut.shape[0]):
        residual_coll.append(Solution_cut[i, 0])
        Solution_args = Solution_cut[i, 1]
        # vsd = np.interp(r_grid_eff, Solution_cut[i, 1], Solution_cut[i, 2][:-2], left=0, right=0)
        vsd = np.interp(ln_r_grid_eff, np.log(Solution_args['r_grid_inv']), Solution_cut[i, 2][:-2], left=0,  right=0)  # both the kernel function and dV_dlnr should be interpolated in the log-scale
        vsd_coll.append(vsd)
        n_coll.append(Solution_cut[i, 2][-2])
        k_coll.append(Solution_cut[i, 2][-1])
        vt, reff = Vt_Reff(grid=Solution_args['r_grid_inv'], vi=Solution_cut[i, 2][:-2])
        vt_coll.append(vt)
        reff_coll.append(reff)
        Solution_args.pop('r_grid_inv')
        Solution_args['only_fwd'] = True
        wave_union, ext_cal, bac_cal, ssa_cal, lr_cal, depol_cal = get_opt(Solution_cut[i, 2], **Solution_args)
        for item, container in zip([ext_cal, bac_cal, ssa_cal, lr_cal, depol_cal],
                                   [ext_full_cal_coll, bac_full_cal_coll, ssa_full_coll, lr_full_cal_coll,
                                    depol_full_cal_coll]):
            container.append(item)
    vsd_mean, vsd_std = statistic_quant(vsd_coll)
    cri_r_mean, cri_r_std = statistic_quant(n_coll)
    cri_i_mean, cri_i_std = statistic_quant(k_coll)
    vt_mean, vt_std = statistic_quant(vt_coll)
    reff_mean, reff_std = statistic_quant(reff_coll)
    ext_full_cal_mean, ext_full_cal_std = statistic_quant(ext_full_cal_coll)
    bac_full_cal_mean, bac_full_cal_std = statistic_quant(bac_full_cal_coll)
    ssa_full_mean, ssa_full_std = statistic_quant(ssa_full_coll)
    depol_full_cal_mean, depol_full_cal_std = statistic_quant(depol_full_cal_coll)
    lr_full_cal_mean, lr_full_cal_std = statistic_quant(lr_full_cal_coll)
    residual_mean, residual_std = statistic_quant(residual_coll)
    vt_eff, reff_eff = Vt_Reff(r_grid_eff, vsd_mean)
    state_param = np.hstack((vsd_mean, cri_r_mean, cri_i_mean))
    #args_for_recal = {'x_array': state_param, 'r_grid': r_grid_eff, 'model': model}
    wave_union, ext_full_cal_eff, bac_full_cal_eff, ssa_full_eff, lr_full_cal_eff, depol_full_cal_eff = get_opt(state_param, r_grid_eff, only_fwd=True, model=model, mI_dusttype=mI_dusttype)
    #y1_cal_eff = get_opt(only_fwd=False, **args_for_recal)
    y1_cal_eff = np.concatenate((ext_full_cal_eff[:len(ext_wave)], bac_full_cal_eff[:len(bac_wave)]))
    if depol_wave is not None:
        y1_cal_eff = np.append(y1_cal_eff, depol_full_cal_eff[:len(depol_wave)])
    residual_eff = np.linalg.norm((y1_cal_eff - y1) / y1) * 100

    #vsd_mean = np.concatenate(([0], vsd_mean, [0]))
    #vsd_std = np.concatenate(([0], vsd_std, [0]))
    #r_grid_eff = r_grid_org[rmin_start - 1: rmax_end + 2]

    retrieval = {'rtv_message': rtv_message, 'rtv_flag': rtv_flag, 'str_time': str_time,
                 'fwd_model': model, 'NIS': Num_of_solutions, 'r_grid': r_grid_eff,
                 'VSD_IS_all': VSD_IS, 'VSD_IS_selected': vsd_coll,
                 'VSD_mean': vsd_mean, 'VSD_std': vsd_std,
                 'CRI_r_mean': cri_r_mean, 'CRI_r_std': cri_r_std,
                 'CRI_i_mean': cri_i_mean, 'CRI_i_std': cri_i_std,
                 'Vt_mean': vt_mean, 'Vt_std': vt_std, 'Vt_eff': vt_eff,
                 'Reff_mean': reff_mean, 'Reff_std': reff_std, 'Reff_eff': reff_eff,
                 'Wave_union': wave_union,
                 'SSA_full_mean': ssa_full_mean, 'SSA_full_std': ssa_full_std, 'SSA_full_eff': ssa_full_eff}
    fit = {'rtv_message': rtv_message, 'str_time': str_time,
           'Ext_meas': ext_series, 'Ext_meas_err': ext_err_series,
           'Bac_meas': bac_series, 'Bac_meas_err': bac_err_series,
           'Depol_meas': depol_series, 'Depol_err': depol_err_series,
           'LR_meas': lr_series, 'LR_err': lr_err_series, 'Wave_union': wave_union,
           'Residual_mean': residual_mean, 'Residual_std': residual_std, 'Residual_eff': residual_eff,
           'Ext_full_cal_mean': ext_full_cal_mean, 'Ext_full_cal_std': ext_full_cal_std,
           'Ext_full_cal_eff': ext_full_cal_eff,
           'Bac_full_cal_mean': bac_full_cal_mean, 'Bac_full_cal_std': bac_full_cal_std,
           'Bac_full_cal_eff': bac_full_cal_eff,
           'Depol_full_cal_mean': depol_full_cal_mean, 'Depol_full_cal_std': depol_full_cal_std,
           'Depol_full_cal_eff': depol_full_cal_eff,
           'LR_full_cal_mean': lr_full_cal_mean, 'LR_full_cal_std': lr_full_cal_std, 'LR_full_cal_eff': lr_full_cal_eff
           }
    return retrieval, fit


def plot_rtv(retrieval, output_dir, name):
    """
    plot retrieval results
    :param retrieval: dict, retrieval results returned by the method inversion()
    :param output_dir: str, prefix of the figure name, indicating the saving directory
    :param name: str, name of the plot
    :return: filepath
    """
    if output_dir is None:
        filepath = name + '_retrieval.png'
    else:
        filepath = os.path.join(output_dir, name + '_retrieval.png')
        if retrieval.keys() == 2:
            print('There is nothing to plot because: ' + retrieval['rtv_message'])
        else:
            # about the table
            tab_rowlabel = ['mean', 'std', 'eff']
            tab_collabel = ['$V_t$', '$R_{eff}$', '$n$', '$\kappa$']
            tab_text = [
                ['%.2f' % retrieval['Vt_mean'], '%.2f' % retrieval['Reff_mean'], '%.2f' % retrieval['CRI_r_mean'],
                 '%.3f' % retrieval['CRI_i_mean']],
                ['%.3f' % retrieval['Vt_std'], '%.3f' % retrieval['Reff_std'], '%.3f' % retrieval['CRI_r_std'],
                 '%.4f' % retrieval['CRI_i_std']],
                ['%.2f' % retrieval['Vt_eff'], '%.2f' % retrieval['Reff_eff'], ' ', ' ']]
            fontsize = 14
            labelsize = 12
            font_label = FontProperties(weight='semibold', size=fontsize, variant='small-caps')
            # process_at = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

            '''now = datetime.datetime.now()
            prs_time = now.strftime('%Y-%m-%dT%H:%M:%S')
            local_now = now.astimezone()
            tz_str = local_now.tzinfo.tzname(local_now)
            plt.figtext(0.8, 0.01, "Processed at : %s (%s time)" % (prs_time, tz_str), ha="center", fontsize=10,
                        color='gray')'''

            fig1 = plt.figure(figsize=(9, 5))  # , constrained_layout=True)
            gs = GridSpec(5, 5, figure=fig1)
            ax_left = fig1.add_subplot(gs[:, :3])
            ax_left.grid()
            ax_upright = fig1.add_subplot(gs[:2, 3:])
            ax_upright.grid()
            ax_upright.grid(which='minor', ls='--')
            ax_midright = fig1.add_subplot(gs[2: 4, 3:])
            ax_lowrigh = fig1.add_subplot(gs[-1, 3:])

            ax_left.errorbar(retrieval['r_grid'], retrieval['VSD_mean'], yerr=retrieval['VSD_std'], marker='o',
                             capsize=3)
            ax_left.tick_params(labelsize=labelsize)
            ax_left.set_xlabel('r $({\mu}m)$', fontsize=fontsize)  # fontproperties=font_label)
            ax_left.set_ylabel('dV/dlnr $({\mu}m^3/cm^3)$', fontsize=fontsize)  # fontproperties=font_label)
            ax_left.set_xscale('log')

            ax_upright.errorbar(retrieval['Wave_union'], retrieval['SSA_full_mean'], yerr=retrieval['SSA_full_std'],
                                marker='o',
                                capsize=3)
            ax_upright.tick_params(labelsize=labelsize)
            ax_upright.set_xlabel('${\lambda}({\mu}m)$', fontsize=fontsize)  # fontproperties=font_label)
            ax_upright.set_ylabel('SSA', fontsize=fontsize)  # fontproperties=font_label)
            ax_upright.yaxis.set_minor_locator(AutoMinorLocator(2))

            ax_tab = ax_midright.table(cellText=tab_text, cellLoc='center', rowLabels=tab_rowlabel,
                                       colLabels=tab_collabel, rowLoc='center', colLoc='center', loc='center',
                                       fontsize=fontsize)
            ax_tab.scale(1, 2)
            ax_midright.axis('off')

            ax_lowrigh.text(0, 0.75, 'Number of IS: %d\n$V_t$ in ${\mu}m^3/cm^3$, $R_{eff}$ in ${\mu}m$' % (
                retrieval['NIS']), fontsize=labelsize, transform=ax_lowrigh.transAxes)
            font_strtime = FontProperties(style='italic')
            ax_lowrigh.text(0, 0, retrieval['str_time'], fontsize=labelsize - 2, transform=ax_lowrigh.transAxes,
                            color='gray', fontproperties=font_strtime)
            ax_lowrigh.set_frame_on(False)
            ax_lowrigh.xaxis.set_visible(False)
            ax_lowrigh.yaxis.set_visible(False)
            fig1.suptitle(name + '_retrieval', fontproperties=font_label)
            plt.subplots_adjust(wspace=1.1, hspace=0.4)
            plt.savefig(filepath, dpi=300)
            plt.close()
    return filepath


def plot_fit(fit, output_dir, name):
    """
    plot fitting result
    :param fit: dict, fitted and recalculated optical properties returned by the method inversion()
    :param output_dir: str, prefix of the figure name, indicating the saving directory
    :param name: str, name of the plot
    :return: filepath
    """
    if output_dir is None:
        filepath = name + '_fit.png'
    else:
        filepath = os.path.join(output_dir, name + '_fit.png')
        if fit.keys() == 2:
            print('There is nothing to plot because: ' + fit['rtv_message'])
        else:
            fontsize = 14
            labelsize = 12
            markersize = 8
            font_label = FontProperties(weight='semibold', size=14, variant='small-caps')
            # process_at = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            ext_wave = fit['Ext_meas'].index
            bac_wave = fit['Bac_meas'].index
            lr_wave = fit['LR_meas'].index

            fig2, ax2 = plt.subplots(2, 2, figsize=(9, 5))  # , constrained_layout=True)
            for axes in ax2.ravel():
                axes.tick_params(labelsize=labelsize)
                axes.set_xlabel('Wavelength $({\mu}m)$', fontsize=fontsize)  # fontproperties=font_label)
                axes.xaxis.set_minor_locator(AutoMinorLocator(2))
                axes.yaxis.set_minor_locator(AutoMinorLocator(2))
                axes.grid()
                axes.grid(which='minor', ls='--')
            ax2[0, 0].errorbar(ext_wave, fit['Ext_meas'].values, yerr=fit['Ext_meas_err'] * fit['Ext_meas'].values,
                               marker='o', color='C0',
                               markersize=markersize, linestyle='', label='Measured', capsize=3)
            ax2[0, 0].errorbar(fit['Wave_union'], fit['Ext_full_cal_mean'], yerr=fit['Ext_full_cal_std'], marker='o',
                               markerfacecolor='none', markersize=markersize, linestyle='', label='Cal. mean',
                               capsize=3, color='C3')
            # ax2[0, 0].plot(fit['Wave_union'], fit['Ext_full_cal_eff'], marker='D', markerfacecolor='none', linestyle='',
            # markersize=markersize, label='Cal. effective')
            ax2[0, 0].set_ylabel(r'${\alpha}(Mm^{-1})$', fontsize=fontsize)  # fontproperties=font_label)
            ax2[0, 0].legend(loc='best', title='Residual: {:.2f}%'.format(fit['Residual_mean']))

            ax2[0, 1].errorbar(bac_wave, fit['Bac_meas'].values, yerr=fit['Bac_meas_err'] * fit['Bac_meas'].values,
                               marker='o', color='C0',
                               markersize=markersize, linestyle='', capsize=3)
            ax2[0, 1].errorbar(fit['Wave_union'], fit['Bac_full_cal_mean'], yerr=fit['Bac_full_cal_std'], marker='o',
                               markerfacecolor='none', markersize=markersize, linestyle='', capsize=3, color='C3')
            # ax2[0, 1].plot(fit['Wave_union'], fit['Bac_full_cal_eff'], marker='D', markerfacecolor='none',
            # markersize=markersize, linestyle='')
            ax2[0, 1].set_ylabel(r'${\beta}((Mm*sr)^{-1})$', fontsize=fontsize)  # fontproperties=font_label)

            ax2[1, 0].errorbar(lr_wave, fit['LR_meas'].values, yerr=fit['LR_err'] * fit['LR_meas'].values, marker='o',
                               linestyle='', color='C0',
                               markersize=markersize, capsize=3)
            ax2[1, 0].errorbar(fit['Wave_union'], fit['LR_full_cal_mean'], yerr=fit['LR_full_cal_std'], marker='o',
                               markerfacecolor='none', markersize=markersize, linestyle='', capsize=3, color='C3')
            # ax2[1, 0].plot(fit['Wave_union'], fit['LR_full_cal_eff'], marker='D', markerfacecolor='none', markersize=markersize, linestyle='')
            ax2[1, 0].set_ylabel('Lidar ratio (sr)', fontsize=fontsize)  # fontproperties=font_label)

            if len(fit['Depol_meas']):
                depol_wave = fit['Depol_meas'].index
                ax2[1, 1].errorbar(depol_wave, fit['Depol_meas'],
                                   yerr=fit['Depol_err'] * fit['Depol_meas'], marker='o',
                                   markersize=markersize, linestyle='', capsize=3, color='C0')
            ax2[1, 1].errorbar(fit['Wave_union'], fit['Depol_full_cal_mean'], yerr=fit['Depol_full_cal_std'],
                               marker='o', color='C3', markersize=markersize, markerfacecolor='none', linestyle='', capsize=3)
            ax2[1, 1].set_ylabel('${\delta}$', fontsize=fontsize)  # fontproperties=font_label)
            fig2.suptitle(name + '_fit', fontproperties=font_label)
            plt.subplots_adjust(wspace=0.3, hspace=0)
            plt.savefig(filepath, dpi=300)
            plt.close()
    return filepath


def txt_header():
    return f"""BOREAL

Data policy: The services and data that you have accessed are provided by the Laboratoire d'Optique Atmospherique, 
a joint research unit of the University of Lille and CNRS. If you utilize the aerosol products for publication purposes, 
we kindly request you to cite the paper by Chang et al., 2022 and acknowledge the contribution of 
"University of Lille/CNRS/Laboratoire d'Optique Atmospherique for BOREAL online processing service." 
Additionally, we encourage you to consider offering co-authorship to the scientists who contributed to the development 
of BOREAL, if their involvement is relevant to your work. Your recognition and collaboration contribute to the 
advancement of scientific research and the acknowledgment of the efforts invested in the development of these 
resources for the community.

"""


def export_txt(retrieval, fit, product_output_dir, filename, header=False, site_info=None, timestamp_str=None):
    """
    export retrieval and fit to two txt files
    :param timestamp_str: str, processing time
    :param site_info: str, information on the measurement site
    :param filename: str, filename
    :param product_output_dir: str, output directory
    :param retrieval: retrieval results returned by the method inversion()
    :param fit: fitted and recalculated optical properties returned by the method inversion()
    :return: two txt files
    """
    if product_output_dir is not None:
        filepath = os.path.join(product_output_dir, filename + '_output.txt')

        f = open(filepath, 'w+')
        if header:
            f.write(txt_header())

        if site_info is not None:
            for ele in site_info:
                f.write("%s = %s\n" % (ele, str(site_info[ele])))

        if timestamp_str is not None:
            timestamp_str = timestamp_str.replace(',', '_')
            timestamp_str = timestamp_str.replace(':', '-')
            f.write("Average Time = %s\n" % timestamp_str)

        f.write('\n' + "*" * 30 + "" + "   Retrievals   " + "*" * 30 + '\n')
        if fit is None:
            f.write(retrieval['rtv_message'] + '\n')
            f.write(retrieval['str_time'])
        else:
            f.write(retrieval['fwd_model'] + ' model is used.\n')
            f.write('Num. of selected individual solutions: {}\n'.format(retrieval['NIS']))
            f.write(
                'The following mean and std are statistical results of the individual solutions, not retrieval uncertainty\n')
            f.write('\tmean\tstd\n')
            f.write('mR\t{:.2f}\t{:.3f}\n'.format(retrieval['CRI_r_mean'], retrieval['CRI_r_std']))
            f.write('mI\t{:.3f}\t{:.4f}\n'.format(retrieval['CRI_i_mean'], retrieval['CRI_i_std']))
            f.write('Vt(um^3/cm^3)\t{:.2f}\t{:.3f}\n'.format(retrieval['Vt_mean'], retrieval['Vt_std']))
            f.write('Reff(um)\t{:.2f}\t{:.3f}\n'.format(retrieval['Reff_mean'], retrieval['Reff_std']))
            f.write('r(um)\tdv/dlnr_mean(um^3/cm^3)\tdv/dlnr_std(um^3/cm^3)\n')
            for r, vsd, vsd_std in zip(retrieval['r_grid'], retrieval['VSD_mean'], retrieval['VSD_std']):
                f.write('{:.5e}\t{:.5e}\t{:.5e}\n'.format(r, vsd, vsd_std))

            f.write('\n' + "*" * 30 + "" + "   Fittings   " + "*" * 30 + '\n')

            wave_list = [str(int(wave * 1000)) for wave in fit['Wave_union']]
            wave_str = 'wave(nm)\t'
            for wave in wave_list:
                wave_str += wave + '\t'
            f.write(wave_str + '\n')
            Meas_list = []
            Cal_list = []
            Cal_std_list = []
            Opt_type = ['Ext', 'Bac', 'LR']
            Opt_unit = ['(1/Mm)', '(1/(Mm*sr))', '(sr)']
            for j, opt in enumerate(Opt_type):
                meas = opt + '_meas' + Opt_unit[j] + '\t'
                cal = opt + '_cal\t'
                cal_std = opt + '_cal_std\t'
                meas_reindex = fit[opt + '_meas'].reindex(fit['Wave_union'], fill_value=-999)
                for i in range(len(wave_list)):
                    meas += '%.5e\t' % meas_reindex.iloc[i]
                    cal += '%.5e\t' % fit[opt + '_full_cal_mean'][i]
                    cal_std += '%.5e\t' % fit[opt + '_full_cal_std'][i]
                Meas_list.append(meas + '\n')
                Cal_list.append(cal + '\n')
                Cal_std_list.append(cal_std + '\n')
            for i in range(len(Opt_type)):
                f.write(Meas_list[i])
                f.write(Cal_list[i])
                f.write(Cal_std_list[i])
            meas = 'Depol_meas\t'
            cal = 'Depol_cal\t'
            cal_std = 'Depol_cal_std\t'
            if len(fit['Depol_meas']):
                meas_depol = fit['Depol_meas'].reindex(fit['Wave_union'], fill_value=-999).values
            else:
                meas_depol = [-999] * len(wave_list)
            for i in range(len(wave_list)):
                meas += '%.5e\t' % meas_depol[i]
                cal += '%.5e\t' % fit['Depol_full_cal_mean'][i]
                cal_std += '%.5e\t' % fit['Depol_full_cal_std'][i]
            f.write(meas + '\n')
            f.write(cal + '\n')
            f.write(cal_std + '\n')
            ssa_cal = 'SSA_cal\t'
            ssa_cal_std = 'SSA_cal_std\t'
            for i in range(len(wave_list)):
                ssa_cal += '%.5e\t' % retrieval['SSA_full_mean'][i]
                ssa_cal_std += '%.5e\t' % retrieval['SSA_full_std'][i]
            f.write(ssa_cal + '\n')
            f.write(ssa_cal_std + '\n')
            f.write(
                'Tot_residual\t{:.2f}\tTot_residual_std\t{:.3f}\n'.format(fit['Residual_mean'], fit['Residual_std']))

            f.write('\n' + "*" * 30 + "" + "   Message   " + "*" * 30 + '\n')
            f.write(retrieval['rtv_message'] + '\n')
            f.write(retrieval['str_time'])
    else:
        filepath = filename + '_output.txt'
    return filepath


# default_config, default_config2 = get_default_confg()
custom_config = {'rmin_min': 0.05, 'rmin_max': 0.3, 'rmax_min': 1.5, 'rmax_max': 15, 'r_regrid_num': 8, 'na': 1.5,
                 'sigma_na': 0.1, 'ka': 0.01, 'sigma_ka': 0.01, 'epsilon2': 2.5, 'epsilon_delta_x': 2.54,
                 'epsilon_delta_n': 0.07, 'epsilon_delta_k': 2.3}



# this module contains functions for using the Irregular-Hexahedral scattering model (TAMUdust2020: Saito et al., 2021)
# use the sub-dataset extracted from TAMUdust-2020 for forward_module calculation

import os
import numpy as np
import math
from scipy.interpolate import interp1d


def __interp_nk(ker, eta, kappa):
    """
    implement 2-dim interpolation to obtain the kernel with specified (eta, kappa)
    :param kappa: float, imaginary part of the CRI
    :param eta: float, Real part of the CRI
    :param ker: kernel matrix to interpolate with shape (mR, mI, r)
    :return:
    """
    ii = max(mR[mR < eta].size - 1, 0)
    jj = max(mI[mI < kappa].size - 1, 0)
    n_left, n_right = mR[ii], mR[ii + 1]
    k_left, k_right = mI[jj], mI[jj + 1]
    func1 = interp1d([math.log(k_left), math.log(k_right)], np.array([ker[ii, jj], ker[ii, jj + 1]]), axis=0)  # linear interp for k on the log scal
    kernel_nleft_k = func1(math.log(kappa))
    func2 = interp1d([math.log(k_left), math.log(k_right)], np.array([ker[ii + 1, jj], ker[ii + 1, jj + 1]]), axis=0)
    kernel_nright_k = func2(math.log(kappa))  # linear interp for k on the log scale
    func3 = interp1d([n_left, n_right], np.array([kernel_nleft_k, kernel_nright_k]), axis=0)
    kernel_eta_kappa = func3(eta)
    return kernel_eta_kappa


'''def get_opt2(x_array, r_grid, only_fwd=False, wave_ext=None, wave_bac=None, wave_depol=None, lnr_grid_redu=None):
    """
        get bulk optical properties for a fixed sphericity.

        :param x_array: ndarray, x[:-2]: particle volume size distribution (VSD) with respect to volume-equivalent radius, r_ve: v(r) = dV/dlnr;
                              x[-2]: mR; x[-1]: mI
        :param r_grid: ndarray, r_ve grid. The first and last elements must coincide with the elements in r_grid_org.
        :param lnr_grid_redu: ndarray, logarithm of reduced radius grids for inversion, r_grid[0] and r_grid[-1] are included. These grids are logarithmic equidistant in [r_grid[0], r_grid[-1]].
                    the length should be equal to the length of dV_dlnr.
        :param wave_bac: list, wavelengths of backscattering
        :param wave_ext: list, wavelengths of extinction
        :param wave_depol: list, wavelength of depolarization
        :param only_fwd: whether it is only used for forward_module calculation. If True, wave + ext + bac + ssa + lr (+ pldr) at full waves will be output;
        otherwise, a ndarray of simulated measurements matching the input [ext, bac(, depol)] will be output for fitting.
        :return:
    """

    v_lnr_ = x_array[:-2]  # load VSD
    lnr_grid = np.log(r_grid)
    ind_n_left = mR[mR < x_array[-2]].size - 1
    ind_k_left = mI[mI < x_array[-1]].size - 1
    lnk = math.log(x_array[-1])
    lnk_left, lnk_right = math.log(mI[ind_k_left]), math.log(mI[ind_k_left + 1])
    n_left, n_right = mR[ind_n_left], mR[ind_n_left + 1]
    interp_fun1 = interp1d([lnk_left, lnk_right], np.array([OptProp_DB2[ind_n_left, ind_k_left], OptProp_DB2[ind_n_left, ind_k_left+1]]), axis=0)
    OptProp_nleft_k = interp_fun1(lnk)
    interp_fun2 = interp1d([lnk_left, lnk_right], np.array([OptProp_DB2[ind_n_left+1, ind_k_left], OptProp_DB2[ind_n_left+1, ind_k_left+1]]), axis=0)
    OptProp_nright_k = interp_fun2(lnk)
    interp_fun3 = interp1d([n_left, n_right], np.array([OptProp_nleft_k, OptProp_nright_k]), axis=0)
    Sub_OptProp_DB = interp_fun3(x_array[-2])

    Sub_OptProp_DB = Sub_OptProp_DB[:, (r_grid_org >= r_grid[0]) & (r_grid_org <= r_grid[-1]), :]

    if lnr_grid_redu is not None:
        v_lnr_interpd = np.interp(lnr_grid, lnr_grid_redu, v_lnr_)
        Sub_OptProp_DB_bulk = np.sum(Sub_OptProp_DB * v_lnr_interpd[:, np.newaxis], axis=1)
    else:
        Sub_OptProp_DB_bulk = np.sum(Sub_OptProp_DB * v_lnr_[:, np.newaxis], axis=1)

    ext = Sub_OptProp_DB_bulk[:, 0]
    bac = Sub_OptProp_DB_bulk[:, 1]
    sca = Sub_OptProp_DB_bulk[:, 2]
    ssa = sca / ext
    LidarRatio = ext / bac
    P11_pi = Sub_OptProp_DB_bulk[:, 3]
    P22_pi = Sub_OptProp_DB_bulk[:, 4]
    pldr = (P11_pi - P22_pi) / (P11_pi + P22_pi)

    if only_fwd:
        wave_full = [0.355, 0.532, 1.064]
        return wave_full, ext, bac, ssa, LidarRatio, pldr
    else:
        ext_ind = [Wave.index(w) for w in wave_ext]
        bac_ind = [Wave.index(w) for w in wave_bac]
        if wave_depol is None:
            opt = np.concatenate((ext[ext_ind], bac[bac_ind]))
        else:
            depol_ind = [Wave.index(w) for w in wave_depol]
            opt = np.concatenate((ext[ext_ind], bac[bac_ind], pldr[depol_ind]))
        return opt'''


def get_opt(x_array, r_grid, only_fwd=False, wave_ext=None, wave_bac=None, wave_depol=None, lnr_grid_redu=None, mI_dusttype=False, model='ih'):
    """
        get bulk optical properties for a fixed sphericity=0.695.
        :param model: reserved argument, in order to keep consistent with the input of spheroid.get_opt
        :param mI_dusttype: if consider mI spectral dependency of the dust type
        :param x_array: ndarray, x[:-2]: particle volume size distribution (VSD) with respect to volume-equivalent radius, r_ve: v(r) = dV/dlnr;
                              x[-2]: mR; x[-1]: mI
        :param r_grid: ndarray, r_ve grid. The first and last elements must coincide with the elements in r_grid_org.
        :param lnr_grid_redu: ndarray, logarithm of reduced radius grids for inversion, r_grid[0] and r_grid[-1] are included. These grids are logarithmic equidistant in [r_grid[0], r_grid[-1]].
                    the length should be equal to the length of dV_dlnr.
        :param wave_bac: list, wavelengths of backscattering
        :param wave_ext: list, wavelengths of extinction
        :param wave_depol: list, wavelength of depolarization
        :param only_fwd: whether it is only used for forward_module calculation. If True, wave + ext + bac + ssa + lr (+ pldr) at full waves will be output;
        otherwise, a ndarray of simulated measurements matching the input [ext, bac(, depol)] will be output for fitting.
        :return:
    """

    v_lnr_ = x_array[:-2]  # load VSD
    lnr_grid = np.log(r_grid)
    OptProp_DB2_bulk = []
    if mI_dusttype:  # dust-type mI
        kappa = np.array([x_array[-1], x_array[-1] * 0.52, 0.001])
    else:
        kappa = np.ones(len(Wave)) * x_array[-1]
    for i in range(len(Wave)):
        Sub_OptProp_DB = __interp_nk(OptProp_DB2[:, :, i, :, :], x_array[-2], kappa[i])
        Sub_OptProp_DB = Sub_OptProp_DB[(r_grid_org >= r_grid[0]) & (r_grid_org <= r_grid[-1])]
        if lnr_grid_redu is not None:
            v_lnr_interpd = np.interp(lnr_grid, lnr_grid_redu, v_lnr_)
            Sub_OptProp_DB_bulk = np.sum(Sub_OptProp_DB * v_lnr_interpd[:, np.newaxis], axis=0)
        else:
            Sub_OptProp_DB_bulk = np.sum(Sub_OptProp_DB * v_lnr_[:, np.newaxis], axis=0)
        OptProp_DB2_bulk.append(Sub_OptProp_DB_bulk)
    OptProp_DB2_bulk = np.array(OptProp_DB2_bulk)

    ext = OptProp_DB2_bulk[:, 0]
    bac = OptProp_DB2_bulk[:, 1]
    sca = OptProp_DB2_bulk[:, 2]
    ssa = sca / ext
    LidarRatio = ext / bac
    P11_pi = OptProp_DB2_bulk[:, 3]
    P22_pi = OptProp_DB2_bulk[:, 4]
    pldr = (P11_pi - P22_pi) / (P11_pi + P22_pi)

    if only_fwd:
        return Wave, ext, bac, ssa, LidarRatio, pldr
    else:
        ext_ind = [Wave.index(w) for w in wave_ext]
        bac_ind = [Wave.index(w) for w in wave_bac]
        if wave_depol is None:
            opt = np.concatenate((ext[ext_ind], bac[bac_ind]))
        else:
            depol_ind = [Wave.index(w) for w in wave_depol]
            opt = np.concatenate((ext[ext_ind], bac[bac_ind], pldr[depol_ind]))
        return opt


'''def get_optdust(x_array, r_grid, wave_ext, wave_bac, wave_depol=None, lnr_grid_redu=None, only_fwd=False):
    """
        get bulk optical properties for a fixed sphericity of 0.71.

        :param x_array: ndarray, x[:-2]: particle volume size distribution (VSD) with respect to volume-equivalent radius, r_ve: v(r) = dV/dlnr;
                              x[-2]: mR; x[-1]: mI
        :param r_grid: ndarray, r_ve grid. The first and last elements must coincide with the elements in r_grid_org.
        :param lnr_grid_redu: ndarray, logarithm of reduced radius grids for inversion, r_grid[0] and r_grid[-1] are included. These grids are logarithmic equidistant in [r_grid[0], r_grid[-1]].
                    the length should be equal to the length of dV_dlnr.
        :param wave_bac: list, wavelengths of backscattering
        :param wave_ext: list, wavelengths of extinction
        :param wave_depol: list, wavelength of depolarization
        :param only_fwd: whether it is only used for forward_module calculation. If True, wave + ext + bac + ssa + lr (+ pldr) at full waves will be output;
        otherwise, a ndarray of simulated measurements matching the input [ext, bac(, depol)] will be output for fitting.
        :return:
    """
    v_lnr_ = x_array[:-2]  # load VSD
    n = x_array[-2]
    lnr_grid = np.log(r_grid)
    OptProp_DB2_bulk = []
    k_list = [x_array[-1], x_array[-1] * 0.52, 0.001]  # the coefficient is derived from regression analysis
    for i in range(len(Wave)):
        k = k_list[i]
        Sub_OptProp_DB = __interp_nk(OptProp_DB2[:, :, i, :, :], n, k)
        Sub_OptProp_DB = Sub_OptProp_DB[(r_grid_org >= r_grid[0]) & (r_grid_org <= r_grid[-1])]
        if lnr_grid_redu is not None:
            v_lnr_interpd = np.interp(lnr_grid, lnr_grid_redu, v_lnr_)
            Sub_OptProp_DB_bulk = np.sum(Sub_OptProp_DB * v_lnr_interpd[:, np.newaxis], axis=0)
        else:
            Sub_OptProp_DB_bulk = np.sum(Sub_OptProp_DB * v_lnr_[:, np.newaxis], axis=0)
        OptProp_DB2_bulk.append(Sub_OptProp_DB_bulk)
    OptProp_DB2_bulk = np.array(OptProp_DB2_bulk)

    ext = OptProp_DB2_bulk[:, 0]
    bac = OptProp_DB2_bulk[:, 1]
    sca = OptProp_DB2_bulk[:, 2]
    ssa = sca / ext
    LidarRatio = ext / bac
    P11_pi = OptProp_DB2_bulk[:, 3]
    P22_pi = OptProp_DB2_bulk[:, 4]
    pldr = (P11_pi - P22_pi) / (P11_pi + P22_pi)

    if only_fwd:
        wave_full = [0.355, 0.532, 1.064]
        return wave_full, ext, bac, ssa, LidarRatio, pldr
    else:
        ext_ind = [Wave.index(w) for w in wave_ext]
        bac_ind = [Wave.index(w) for w in wave_bac]
        if wave_depol is None:
            opt = np.concatenate((ext[ext_ind], bac[bac_ind]))
        else:
            depol_ind = [Wave.index(w) for w in wave_depol]
            opt = np.concatenate((ext[ext_ind], bac[bac_ind], pldr[depol_ind]))
        return opt'''


# read sub-database
current_dir = os.path.dirname(os.path.abspath(__file__))
data_path = os.path.join(current_dir, 'data')
sub_DB = np.load(os.path.join(data_path, 'OptProp_new.npy'))  # columns: ['wave', 'D', 'V', 'A', 'Q_ext', 'SSA', 'g', 'P11', 'P22', 'r_ve']
mR = np.arange(1.38, 1.71, 0.02).round(2)
mI = np.array([1.000000e-04, 5.000000e-04, 8.189469e-04, 1.341348e-03,
               2.196985e-03, 3.598428e-03, 5.893843e-03, 9.653489e-03,
               1.581139e-02, 2.589737e-02, 4.241714e-02, 6.947477e-02,
               0.1]).round(4)
Wave = [0.355, 0.532, 1.064]
r_grid_org = sub_DB[: 41, -1].round(4)
d_grid_org = sub_DB[: 41, 1].round(4)

OptProp_DB2 = np.load(os.path.join(data_path, 'OptProp2_Phi695.npy'))  # [mR, mI, wl, r_grid, opt_props(K_ext, K_bac, K_sca, K_P11_pi, K_P22_pi)]



# Retrieval of the state parameters: [Vt, rVc, Sgc, fvf, mR, mI355] from 3+2+3 lidar measurements

import os
import numpy as np
from scipy.optimize import least_squares
from .forward_module.tamuDUST import ih
#import forward_module.tamuDUST.ih as fwd_model
from pandas import DataFrame, Series

fwd_model = ih


def combined_SD(cov_x, grad_x):
    """
    calculate the combined uncertainty of a scalar y which is a function of multivariate x
    :param cov_x: covariance matrix of the vector x
    :param grad_x: gradient vector of y to x
    :return:
    """
    var_y = 0
    for ii in range(cov_x.shape[0]):
        for jj in range(cov_x.shape[1]):
            var_y += cov_x[ii, jj] * grad_x[0, ii] * grad_x[0, jj]
    return var_y**0.5


def derivative(func, xx, args=None, dx=1e-7):
    """
    calculate the Jacobian matrix and the value vector of the vector-function 'f' at 'x'
    :param args: other parameters in f(x), a dictionary
    :param func: objective function, callable, output a 1-D array as the objective vector
    :param xx: the variety vector of 'f', 1-D array (ndarray or list)
    :param dx: delta
    :return: J(x)-ndarray and f(x)-ndarray
    """
    if args is None:
        args = {}
    xx = list(xx)
    y = func(xx, **args)
    try:
        Jacob_T = np.empty((len(xx), len(y)))
    except TypeError:
        Jacob_T = np.empty((len(xx), 1))
    for line_num in range(Jacob_T.shape[0]):
        x_new = xx[:]
        x_new[line_num] += dx
        y_new = func(x_new, **args)
        Jacob_T[line_num] = (y_new - y) / dx
    Jacob = Jacob_T.T
    return Jacob, y


def export_txt(boreal_pc_rtv, product_output_dir, filename, site_info=None, timestamp_str=None):
    if product_output_dir is not None:
        filepath = os.path.join(product_output_dir, filename + '_output.txt')

        with open(filepath, 'w') as f:

            if site_info is not None:
                for ele in site_info:
                    f.write("%s = %s\n" % (ele, str(site_info[ele])))

            if timestamp_str is not None:
                timestamp_str = timestamp_str.replace(',', '_')
                timestamp_str = timestamp_str.replace(':', '-')
                f.write("Average Time = %s\n" % timestamp_str)

            f.write('\n' + "*" * 20 + "   Retrieval message: %s   " % boreal_pc_rtv['rtv_message'] + "*" * 20 + '\n')
            f.write('Retrieved microphysical properties and SSA:\n')
        boreal_pc_rtv['state_params_rtv'].to_csv(filepath, mode='a')
        with open(filepath, 'a') as f:
            f.write('\nRetrieved volume size distribution:\n')
        boreal_pc_rtv['vsd_rtv'].to_csv(filepath, index_label='r (um)', mode='a')
        with open(filepath, 'a') as f:
            f.write('\nCovariance matrices of the logarithmic state parameters:\n')
        for key, longname in zip(['rtv_cov_tot', 'rtv_cov_se', 'rtv_cov_mn'],
                                 ['Total retrieval covariance', 'Retrieval covariance for smoothing error', 'Retrieval covariance for measurement noise']):
            f = open(filepath, 'a')
            f.write('\n' + longname + '\n')
            f.close()
            boreal_pc_rtv[key].to_csv(filepath, mode='a')
        with open(filepath, 'a') as f:
            f.write('\nOptical measurement fit:\n')
        boreal_pc_rtv['opt_fit'].to_csv(filepath, mode='a')
        with open(filepath, 'a') as f:
            f.write('\nrmse_fit (%%): %.2f' % boreal_pc_rtv['rmse_fit'])
    else:
        filepath = filename + '_output.txt'
    return filepath


def from_Lognormal_to_normalLog(mean, std):
    """
    convert the mean and std of a lognormal distribution in linear domain to the mean and std of its logarithm in the log domain
    :param mean:
    :param std:
    :return:
    """
    rel_std = std / mean
    std_in_log = np.log(rel_std**2 + 1)**0.5
    mean_in_log = 0.5 * np.log(mean**2 / (rel_std**2 + 1))
    return mean_in_log, std_in_log


def fwd_func(ln_state):
    """
    forward function to calculate optical properties from the log of the state x
    :param ln_state: log of 'Vt', 'rVc', 'Sgc', 'fvf', 'mR', 'mI355'
    :return:
    """
    x = np.exp(ln_state)
    vsd = lognormal_sd(radius_grid, rVf, x[3]*x[0], Sgf) + lognormal_sd(radius_grid, x[1], (1-x[3])*x[0], x[2])
    y = fwd_model.get_opt(np.concatenate((vsd, x[4:])), radius_grid, wave_ext=waves[:2], wave_bac=waves, wave_depol=waves, only_fwd=False, mI_dusttype=True)
    return np.log(y)


def get_covrtv_degnonlinear_avemat(cov_mm, cov_a, jac_mat):
    """
    calculate retrieval covariance matrix, degree of nonlinearity and averaging matrix for a specific state
    :return:
    """
    cov_a_I = np.linalg.inv(cov_a)
    cov_m_I = np.linalg.inv(cov_mm)
    KTSmIK = np.matmul(jac_mat.T, np.matmul(cov_m_I, jac_mat))
    cov_rtv = np.linalg.inv(KTSmIK + cov_a_I)
    cov_rtv_noise = np.matmul(cov_rtv, np.matmul(KTSmIK, cov_rtv))
    cov_rtv_a = cov_rtv - cov_rtv_noise
    A_matrix = np.matmul(cov_rtv, KTSmIK)
    return cov_rtv, cov_rtv_a, cov_rtv_noise, A_matrix


def get_integ_params(r_grid, dv_lnr):
    """
    calculate vt, r_mean, Sg, and reff for a given dv_lnr with r_grid
    :param r_grid:
    :param dv_lnr:
    :return:
    """
    lnr_grid = np.log(r_grid)
    vt = np.trapz(dv_lnr, lnr_grid)
    reff = vt / np.trapz(dv_lnr/r_grid, lnr_grid)
    lnr_mean = np.trapz(y=lnr_grid * dv_lnr, x=lnr_grid) / vt
    r_mean = np.exp(lnr_mean)
    lnSg = (np.trapz((lnr_grid - lnr_mean) ** 2 * dv_lnr, lnr_grid) / vt)**0.5
    Sg = np.exp(lnSg)  # geometric standard deviation
    return vt, r_mean, Sg, reff


def get_ssa(ln_state):
    """
    calculate spectral ssa from the ln_state
    :param ln_state: log of Vt, rVc, Sgc, fvf, mR, mI355
    :return:
    """
    state_params = np.exp(ln_state)
    dV_dlnr = lognormal_sd(radius_grid, rVf, state_params[0] * state_params[3], Sgf) + lognormal_sd(radius_grid, state_params[1], (1 - state_params[3]) * state_params[0], state_params[2])
    opt_cal = fwd_model.get_opt(np.concatenate((dV_dlnr, state_params[-2:])), radius_grid, only_fwd=True, mI_dusttype=True)
    return opt_cal[3]


def harmonize_opt_format(ext_dict, bac_dict, depol_dict):
    """
    convert the format of input optical data to the standard input of BOREAL-PC
    :param ext_dict: dict, extinction at 355 and 532 nm
    :param bac_dict: dict, backscattering at 355, 532 and 1064 nm
    :param depol_dict: dict, depolarization ratio at 355, 532 and 1064 nm
    :return:
    """
    #meas_array = np.array([ext_dict['355'], ext_dict['532'], bac_dict['355'], bac_dict['532'], bac_dict['1064'],
                          #depol_dict['355'], depol_dict['532'], depol_dict['1064']])
    meas_dict = {}
    for keys, dicts in zip(['ext_', 'bac_', 'depol_'], [ext_dict, bac_dict, depol_dict]):
        for k, item in dicts.items():
            meas_dict[keys+k] = item
    return Series(meas_dict)


def lognormal_sd(r_grid, r_mod, C, gsd):
    """
    get lognormal size distribution (dC/dlnr) for a set of radius grids
    :param r_grid: 1D-ndarray, radius grids in um
    :param r_mod: float, median or mode radius
    :param C: float, concentration
    :param gsd: float, geometry standard deviation
    :return: array of the size distribution having the same size as r_grid
    """
    ln_Sg = np.log(gsd)
    return C * np.exp(-(np.log(r_grid) - np.log(r_mod)) ** 2 / (2 * ln_Sg ** 2)) / ((2 * np.pi) ** 0.5 * ln_Sg)


def reff_from_logstate(ln_state):
    """
    calculate effective radius from the log of the state vector.
    :param ln_state: logarithm of [rVc, Sgc, fvf]
    :return:
    """
    rVc, Sgc, fvf = np.exp(ln_state)
    M2 = np.array([rV**2 * np.exp(-4*np.log(Sg)**2) for rV, Sg in zip([rVf, rVc], [Sgf, Sgc])])
    M3 = np.array([rV**3 * np.exp(-4.5*np.log(Sg)**2) for rV, Sg in zip([rVf, rVc], [Sgf, Sgc])])
    Vt_weight = np.array([fvf, (1 - fvf)])
    return np.sum(M3 * Vt_weight) / np.sum(M2 * Vt_weight)


def resi_vec(ln_state_vector, measurement_vector, meas_ind, std_of_ln_meas_valid_data, apriori_value=None, apriori_std=None):
    """
    residual vector to minimize
    :param meas_ind: keys corresponding to the measurement_vector (to specify the order of the opt params)
    :param std_of_ln_meas_valid_data: pure data vector of the std_of_ln_meas Series
    :param ln_state_vector: ln[Vt, rVc, Sgc, fvf, mR, mI355]
    #:param measurement_vector: 2a+3b+3d lidar measurements
    :param measurement_vector: pure data vector from the measurement Series
    :param apriori_value: if incorporating a priori value
    :param apriori_std: if incorporation a priori standard deviation
    :return:
    """
    ln_opt_cal = fwd_func(ln_state_vector)
    ln_opt_cal_series = Series(ln_opt_cal, index=opt_params_index_full)
    ln_opt_cal_valid_data = ln_opt_cal_series[meas_ind].values
    resi_meas = (np.log(measurement_vector) - ln_opt_cal_valid_data) / std_of_ln_meas_valid_data
    if apriori_value is not None:
        resi_apriori = (ln_state_vector - apriori_value) / apriori_std
        residual_vec = np.concatenate((resi_meas, resi_apriori))
    else:
        residual_vec = resi_meas
    return residual_vec


# configuration:
rVf, Sgf = 0.15, 1.9
waves = [0.355, 0.532, 1.064]
radius_grid = fwd_model.r_grid_org
vt_std = 0.26  # according to the fitting of the in situ measurements
mR_a, mI355_a = 1.518, 0.003
mR_std, mI355_std = 0.022, 0.002
ln_state_param_names = ['ln(Vt)', 'ln(Rc)', 'ln(Sc)', 'ln(fvf)', 'ln(mR)', 'ln(mI355)']

# measurement error
opt_params_index_full = ['ext_355', 'ext_532', 'bac_355', 'bac_532', 'bac_1064', 'depol_355', 'depol_532', 'depol_1064']
max_meas_err = Series([0.1, 0.1, 0.1, 0.1, 0.2, 0.15, 0.15, 0.15], index=opt_params_index_full)
rel_std_of_meas = max_meas_err / 3
std_of_ln_meas = (np.log(rel_std_of_meas**2 + 1))**0.5


class Retrieval_bimodal:
    def __init__(self, opt_meas):  # opt_meas is a pd.Series with keys in the order 'ext_wave', 'bac_wave', 'depol_wave'
        self.opt_meas_data = opt_meas.values
        self.opt_meas_ind = opt_meas.index
        # determine the a priori constriant on vt using the linear regression result with ext_532
        vt_a = 0.89 * opt_meas['ext_532']
        rV_a, Sg_a = 2.97, 1.96
        rV_std, Sg_std = 1.12, 0.34
        fvf_a, fvf_std = 0.028, 0.045
        self.state_a, self.state_std_a = from_Lognormal_to_normalLog(np.array([vt_a, rV_a, Sg_a, fvf_a, mR_a, mI355_a]), np.array([vt_std * vt_a, rV_std, Sg_std, fvf_std, mR_std, mI355_std]))
        self.bounds = ([-np.inf, -np.inf, self.state_a[2] - 3*self.state_std_a[2], self.state_a[3] - 3*self.state_std_a[3], np.log(1.4), np.log(0.0001)], [np.inf, np.inf, self.state_a[2] + 3*self.state_std_a[2], self.state_a[3] + 3*self.state_std_a[3], np.log(1.68), np.log(0.06)])
        self.std_of_ln_meas = std_of_ln_meas[self.opt_meas_ind]
        self.std_of_ln_meas_data = self.std_of_ln_meas.values
        self.cov_m = np.diag(self.std_of_ln_meas_data**2)

    def do_retrieval(self, apriori_constraint=True):
        if apriori_constraint:
            args = [self.opt_meas_data, self.opt_meas_ind, self.std_of_ln_meas, self.state_a, self.state_std_a]
        else:
            args = [self.opt_meas_data, self.opt_meas_ind, self.std_of_ln_meas]
        res = least_squares(resi_vec, self.state_a, bounds=self.bounds, args=args)
        state_rtv = np.exp(res.x)
        rtv_message = res.message
        vsd_rtv = lognormal_sd(radius_grid, rVf, state_rtv[3] * state_rtv[0], Sgf) + \
                  lognormal_sd(radius_grid, state_rtv[1], (1 - state_rtv[3]) * state_rtv[0], state_rtv[2])
        grad_reff_to_lnstate, reff_rtv = derivative(reff_from_logstate, res.x[1:4])
        jac_mat_ssa_to_lnstate, ssa_rtv = derivative(get_ssa, res.x)
        # retrieval covariance and combined uncertainty
        jac_mat_lnopt_to_lnstate, ln_opt = derivative(fwd_func, res.x)
        jac_mat_lnopt_to_lnstate_df = DataFrame(jac_mat_lnopt_to_lnstate, index=opt_params_index_full)
        jac_mat_lnopt_to_lnstate_matter = jac_mat_lnopt_to_lnstate_df.loc[self.opt_meas_ind].values
        cov_a = np.diag(self.state_std_a**2)
        cov_rtv, cov_rtv_smooth_err, cov_rtv_meas_noise, A_mat = get_covrtv_degnonlinear_avemat(self.cov_m, cov_a, jac_mat_lnopt_to_lnstate_matter)
        # standard deviations (absolute) of the retrieved states in the linear space: total std, smoothing error std, meas noise std
        std_rtv_linearspace, reff_std, ssa_std = [], [], []
        for cov in [cov_rtv, cov_rtv_smooth_err, cov_rtv_meas_noise]:
            std_rtv_linearspace.append(((np.exp(np.diag(cov)) - 1) * state_rtv ** 2) ** 0.5)
            reff_std.append(combined_SD(cov[1:4, 1:4], grad_reff_to_lnstate))
            ssa_std.append([combined_SD(cov, jac_mat_ssa_to_lnstate[i, np.newaxis]) for i in range(jac_mat_ssa_to_lnstate.shape[0])])
        # fitting error and recalculated optical data
        residual_vector = res.fun[:len(self.opt_meas_data)]
        opt_recal = self.opt_meas_data / np.exp(residual_vector * self.std_of_ln_meas_data)
        opt_diff = (np.exp(-(residual_vector * self.std_of_ln_meas_data)) - 1) * 100
        fit_rmse = np.linalg.norm(opt_diff) / len(opt_diff) ** 0.5

        # organize the output
        rtv_sd_all = np.concatenate((std_rtv_linearspace, np.array([reff_std]).T, ssa_std), axis=1)
        rtv_result_all = np.concatenate((state_rtv, [reff_rtv], ssa_rtv))
        rtv_all_df = DataFrame(np.vstack((rtv_result_all, rtv_sd_all)).T, index=['Vt', 'Rc', 'Sc', 'fvf', 'mR', 'mI355', 'reff', 'SSA355', 'SSA532', 'SSA1064'],
                               columns=['rtv_value', 'rtv_sd_tot', 'rtv_sd_se', 'rtv_sd_mn'])
        vsd_rtv_series = Series(vsd_rtv, index=radius_grid.round(4).astype(str), name='dV/dlnr (um^3/cm^3)')
        opt_fit_df = DataFrame(np.vstack((self.opt_meas_data, opt_recal, opt_diff)).T, columns=['Opt. measured', 'Opt. fitted', 'Fitting error (%)'],
                               index=self.opt_meas_ind)

        cov_rtv_df = [DataFrame(cov, columns=ln_state_param_names, index=ln_state_param_names) for cov in [cov_rtv, cov_rtv_smooth_err, cov_rtv_meas_noise]]

        return {'rtv_message': rtv_message, 'state_params_rtv': rtv_all_df, 'vsd_rtv': vsd_rtv_series, 'opt_fit': opt_fit_df,
                'rmse_fit': fit_rmse, 'rtv_cov_tot': cov_rtv_df[0], 'rtv_cov_se': cov_rtv_df[1], 'rtv_cov_mn': cov_rtv_df[2]}



# this module contains functions for using the Spheroid scattering model (Dubovik et al. 2006)
import os
import numpy as np
import math
from scipy.interpolate import interp1d


class BOREAL_FwdModelException(Exception):
    pass


def __interp_nk(ker, eta, kappa):
    """
    implement 2-dim interpolation to obtain the kernel with specified (eta, kappa)
    :param kappa: float, imaginary part of the CRI
    :param eta: float, Real part of the CRI
    :param ker: kernel matrix to interpolate with shape (mR, mI, r)
    :return:
    """
    ii = max(mR[mR < eta].size - 1, 0)
    jj = max(mI[mI < kappa].size - 1, 0)
    n_left, n_right = mR[ii], mR[ii + 1]
    k_left, k_right = mI[jj], mI[jj + 1]
    left_kappa = math.log(k_left)   # log interpolate in kappa grid
    right_kappa = math.log(k_right)
    x_kappa = math.log(kappa)
    func1 = interp1d([left_kappa, right_kappa], np.array([ker[ii, jj], ker[ii, jj + 1]]), axis=0)
    kernel_nleft_k = func1(x_kappa)
    func2 = interp1d([left_kappa, right_kappa], np.array([ker[ii + 1, jj], ker[ii + 1, jj + 1]]), axis=0)
    kernel_nright_k = func2(x_kappa)  # linear interp for k on the log scale
    func3 = interp1d([n_left, n_right], np.array([kernel_nleft_k, kernel_nright_k]), axis=0)
    kernel_eta_kappa = func3(eta)
    return kernel_eta_kappa


def __get_Kernel_matrix(eta, kappa, r_grid, wave_list, kernel):
    K_matrix = []
    if isinstance(kappa, float):
        kernel_nk_ = __interp_nk(kernel, eta, kappa)
        for wl in wave_list:
            r_equiv = r_grid * 0.34 / wl
            ln_r_equiv = np.log(r_equiv)
            kernel_equi = 1000 * (0.34 / wl) * np.interp(ln_r_equiv, ln_r_grid_org, kernel_nk_)  # log interpolation in radius grid
            K_matrix.append(kernel_equi)
    else:
        for i in range(len(wave_list)):
            kernel_nk_ = __interp_nk(kernel, eta, kappa[i])
            r_equiv = r_grid * 0.34 / wave_list[i]
            ln_r_equiv = np.log(r_equiv)
            kernel_equi = 1000 * (0.34 / wave_list[i]) * np.interp(ln_r_equiv, ln_r_grid_org, kernel_nk_)  # log interpolation in radius grid
            K_matrix.append(kernel_equi)
    K_matrix = np.mat(np.array(K_matrix))

    return K_matrix


def get_opt(x_array, r_grid, only_fwd=False, wave_ext=None, wave_bac=None, wave_depol=None, model='sphere', lnr_grid_redu=None, mI_dusttype=False):
    """
    the forward model to generate optical properties from micro-physical properties
    :param mI_dusttype: if consider mI spectral dependency of the dust type
    :param wave_bac: ndarray, wavelengths of backscattering
    :param wave_ext: ndarray, wavelengths of extinction
    :param wave_depol: list, wavelength of depolarization
    :param model: str, 'sphere' or 'spheroid', use sphere kernels ('sphere') or spheroid kernels ('spheroid')
    :param only_fwd: whether it is only used for forward calculation. If True, wave + ext + bac + ssa + lr (+ pldr) will be output;
    otherwise, ext + bac will be output for inversion.
    :param r_grid: ndarray, full-resolved grids
    :param lnr_grid_redu: ndarray, logarithm of reduced radius grids for inversion, r_grid[0] and r_grid[-1] are included. These grids are logarithmic equidistant in [r_grid[0], r_grid[-1]].
                    the length should be equal to the length of dV_dlnr.
    :param x_array: list or ndarray, values of each size bin, with the refractive index (x[-2:])
    :return: ndarray, fit optical data (3 beta + 2 alpha)
    """
    vsd = x_array[:-2]
    ln_r_grid = np.log(r_grid)  # both the kernel function and dV_dlnr should be interpolated in the log-scale
    if lnr_grid_redu is not None:
        # Kernel_matrix = Kernel_matrix * np.mat(rep_max)
        vsd_mat = np.mat(np.interp(ln_r_grid, lnr_grid_redu, vsd)).T
    else:
        vsd_mat = np.mat(vsd).T
    wave_full = [0.355, 0.532, 1.064]
    if only_fwd:  # only perform forward calculation, this will output ext_coe, bac_coe, ssa, lr and pldr in the full spectral range
        Opt = []
        if model == 'sphere':
            Kernel_list = [AKernel, BKernel, Abs_Kernel]
        else:  # model == 'spheroid'
            Kernel_list = [AKernel_299, BKernel_299, Abs_Kernel_299, BKernel_p22_299]
        if mI_dusttype:  # dust-type mI
            kappa = np.array([x_array[-1], x_array[-1] * 0.52, 0.001])
        else:  # non-dust-type mI
            kappa = x_array[-1]
        for kernel in Kernel_list:
            kernel_mat = __get_Kernel_matrix(eta=x_array[-2], kappa=kappa, r_grid=r_grid, wave_list=wave_full, kernel=kernel)
            Opt.append(np.array(kernel_mat * vsd_mat).ravel())
        ext = Opt[0]
        bac = Opt[1]
        lr = ext / bac
        absorb = Opt[2]
        ssa = (ext - absorb) / ext
        if model == 'sphere':
            depol = np.zeros(len(wave_full))
        else:
            depol = (bac - Opt[3]) / (bac + Opt[3])
        return wave_full, ext, bac, ssa, lr, depol

    else:  # the output is followed by measurement fitting
        # first: ext and bac
        if model == 'sphere':
            ext_Kernel = AKernel
            bac_Kernel = BKernel
            depol_flag = 0
        else:  # motel == 'spheroid'
            ext_Kernel = AKernel_299
            bac_Kernel = BKernel_299
            depol_flag = 1

        kwargs = {'eta': x_array[-2], 'r_grid': r_grid}
        if mI_dusttype:  # dust-type mI
            kappa = np.array([x_array[-1], x_array[-1] * 0.52, 0.001])
            kappa_ext = kappa[[wave_full.index(i) for i in wave_ext]]
            kappa_bac = kappa[[wave_full.index(i) for i in wave_bac]]
        else:  # non-dust-type mI
            kappa_ext, kappa_bac = x_array[-1], x_array[-1]
        ext_and_bac = []
        for kappa, wave, kernel in zip([kappa_ext, kappa_bac], [wave_ext, wave_bac], [ext_Kernel, bac_Kernel]):
            kernel_mat_ext_or_bac = __get_Kernel_matrix(kappa=kappa, wave_list=wave, kernel=kernel, **kwargs)
            ext_and_bac.append(np.array(kernel_mat_ext_or_bac * vsd_mat).ravel())
        if depol_flag and (wave_depol is not None):
            bulk_p11_and_p11 = []
            if mI_dusttype:  # dust-type mI
                kappa = np.array([x_array[-1], x_array[-1] * 0.52, 0.001])
                kappa_depol = kappa[[wave_full.index(i) for i in wave_depol]]
            else:  # non-dust-type mI
                kappa_depol = x_array[-1]
            for kernel_for_depol in [BKernel_299, BKernel_p22_299]:
                kernel_mat_for_depol = __get_Kernel_matrix(kappa=kappa_depol, wave_list=wave_depol, kernel=kernel_for_depol, **kwargs)
                bulk_p11_and_p11.append(np.array(kernel_mat_for_depol * vsd_mat).ravel())
            depol = (bulk_p11_and_p11[0] - bulk_p11_and_p11[1]) / (bulk_p11_and_p11[0] + bulk_p11_and_p11[1])
            opt = np.concatenate((np.concatenate(ext_and_bac), depol))
        else:
            opt = np.concatenate(ext_and_bac)
        return opt


# load grid refractive index, extinction and back scattering kernels
current_dir = os.path.dirname(os.path.abspath(__file__))
data_path = os.path.join(current_dir, 'data')
mR = np.load(os.path.join(data_path, 'mR.npy'))
mI = np.load(os.path.join(data_path, 'mI.npy'))
AKernel = np.load(os.path.join(data_path, 'AKernel.npy'))
BKernel = np.load(os.path.join(data_path, 'BKernel.npy'))
Abs_Kernel = np.load(os.path.join(data_path, 'Abs_Kernel.npy'))
AKernel_299 = np.load(os.path.join(data_path, 'AKernel_299.npy'))
BKernel_299 = np.load(os.path.join(data_path, 'BKernel_299.npy'))
Abs_Kernel_299 = np.load(os.path.join(data_path, 'Abs_Kernel_299.npy'))
BKernel_p22_299 = np.load(os.path.join(data_path, 'BKernel_p22_299.npy'))
r_grid_org = np.load(os.path.join(data_path, 'r_grid.npy'))
# r_grid_eff = r_grid_org[11: -2]
ln_r_grid_org = np.log(r_grid_org)
